structure Distribution :> Distribution =
struct

  type value = Interpreter.value
 
  type pValue = value * real

  exception DistribError of string*Syntax.pos

  val maxiterations = ref 12


  (* merge two sorted lists of integers *)

  fun mergeI [] l2 = l2
    | mergeI l1 [] = l1
    | mergeI (l1 as (a::l11)) (l2 as (b::l22)) =
        if a<=b then a :: mergeI l11 l2
	else b :: mergeI l1 l22

  fun orderList [] [] = EQUAL
    | orderList [] _ = LESS
    | orderList (a::l1) [] = GREATER
    | orderList (a::l1) (b::l2) =
        if a=b then orderList l1 l2
        else if a<b then LESS
        else GREATER

  (* look up name in list *)

  fun lookup x [] = NONE
    | lookup x ((y,v)::table) =
        if x=y then SOME v else lookup x table

  (* datatype for unnormalized distributions *)

  datatype dist = VAL of value
		| CHOICE of real * dist * dist
		| CPAIR of dist * dist

  fun printD (VAL v) = "VAL [" ^ printL v
    | printD (CHOICE (p,d1,d2)) = "CHOICE(" ^ Real.toString p ^ ","
				  ^ printD d1 ^ "," ^ printD d2 ^ ")"
    | printD (CPAIR (d1,d2)) =  "CPAIR(" ^ printD d1 ^ ","^ printD d2 ^ ")"

  and printL [] = "]"
    | printL [a] = Int.toString a ^ "]"
    | printL (a::l) = Int.toString a ^ "," ^ printL l

  fun printDist d = (TextIO.output (TextIO.stdErr,printD d ^ "\n"); d)

  fun choice (1.0, d1, d2) = d1
    | choice (0.0, d1, d2) = d2
    | choice (p,d1 as (VAL v), d2 as (VAL w)) =
        (case orderList v w of
	   EQUAL => d1
	 | LESS  => CHOICE (p, d1, d2)
	 | _     => CHOICE (1.0-p, d2, d1))
    | choice (p,d1 as (VAL v), d2 as (CHOICE (p1, d3 as (VAL w), d4))) =
        (case orderList v w of
	   EQUAL   => choice (p+p1-p*p1, d1, d4)
	 | GREATER => choice (p1-p*p1, d3,
			 choice (p/(1.0+p*p1-p1), d1, d4))
	 | _       => CHOICE (p, d1, d2))
    | choice (p, d as (CHOICE (p1,d1,d2)), d3) =
	 if d=d3 then d
	 else choice (p*p1, d1, choice ((p-p*p1)/(1.0-p*p1), d2, d3))
    | choice (p, d11 as (CPAIR (d1,d2)),
	         d22 as (CHOICE (p2, CPAIR (d3,d4), d5))) =
	 if d1=d3 then
	   choice (p+p2-p*p2,
		   CPAIR (choice (p/(p+p2-p*p2), d2, d4), d1),
		   d5)
	 else if d1=d4 then
	   choice (p+p2-p*p2,
		   CPAIR (choice (p/(p+p2-p*p2), d2, d3), d1),
		   d5)
	 else if d2=d3 then
	   choice (p+p2-p*p2,
		   CPAIR (choice (p/(p+p2-p*p2), d1, d4), d2),
		   d5)
	 else if d2=d4 then
	   choice (p+p2-p*p2,
		   CPAIR (choice (p/(p+p2-p*p2), d1, d3), d2),
		   d5)
	 else CHOICE (p, d11, d22)
    | choice (p, d1, d as (CHOICE (p1,d2,d3))) =
        if d1=d then d1
        else if d1=d2 then choice (p+p1-p*p1, d1, d3)
	else CHOICE (p,d1,d)
    | choice (p, d1, d2) =
	if d1=d2 then d1
	else CHOICE (p, d1, d2)

  fun cpair (VAL [],d)     = d
    | cpair (d, VAL [])    = d
    | cpair (VAL v, VAL w) = VAL (mergeI v w)
    | cpair (d1, d2)       = CPAIR (d1,d2)

  (* result of normalize is right-associated and sorted CHOICE of VALs *)

  fun normalize (d as (VAL v)) = d
    | normalize (CHOICE (p,d1, d2)) =
        let
	  val d11 = normalize d1
	  val d21 = normalize d2
	in
	  mergeD p d11 d21
	end
    | normalize (CPAIR (d1,d2)) =
        let
	  val d11 = normalize d1
	  val d21 = normalize d2
	in
	  concD d11 d21
	end

  and concD (VAL v) (VAL w) = VAL (mergeI v w)
    | concD (CHOICE (p,d1,d2)) d =
        mergeD p (concD d1 d) (concD d2 d)
    | concD d (CHOICE (p,d1,d2))=
        mergeD p (concD d d1) (concD d d2)
    | concD d1 d2 = (printDist d1; printDist d2;
		     raise DistribError ("Internal error concD",(0,0)))

  and mergeD p (d1 as (VAL v)) (d2 as (VAL w)) =
        (case orderList v w of
	   EQUAL => d1
	 | LESS  => CHOICE (p,d1,d2)
	 | _     => CHOICE (1.0-p,d2,d1))
    | mergeD p (d1 as (VAL v)) (d2 as (CHOICE (p1,d3 as (VAL w),d4))) =
        (case orderList v w of
	   EQUAL => CHOICE (p+p1-p*p1, d1, d4)
	 | LESS  => CHOICE (p,d1,d2)
	 | _     => let
		      val q = (1.0-p)*p1
		    in
		      CHOICE (q, d3,
			      mergeD (p/(1.0-q)) d1 d4)
		    end)
    | mergeD p d1 (d2 as (VAL w)) = mergeD (1.0-p) d2 d1
    | mergeD p (d1 as (CHOICE (p1,d2 as (VAL v),d3)))
	       (d4 as (CHOICE (p2,d5 as (VAL w),d6))) =
        (case orderList v w of
	   EQUAL => let
		      val q = p*p1+(1.0-p)*p2
		    in
		      CHOICE (q, d2,
			      mergeD (p*(1.0-p1)/(1.0-q)) d3 d6)
		    end
	 | LESS  => CHOICE (p*p1, d2,
			    mergeD (p*(1.0-p1)/(1.0-p*p1)) d3 d4)
	 | _     => mergeD (1.0-p) d4 d1)
    | mergeD _ _ _ = raise DistribError ("Internal error mergeD",(0,0))

  fun foldD (VAL v) f g = f v
    | foldD (CHOICE (p,d1,d2)) f g =
        mergeD p (foldD d1 f g) (foldD d2 f g)
    | foldD (CPAIR (d1,d2)) f g =
        (foldUp (foldD d1 f g) (foldD d2 f g) g)

  and foldUp (VAL v) (VAL w) g = g (v, w)
    | foldUp (VAL v) (CHOICE (p,d1,d2)) g =
        mergeD p (foldUp (VAL v) d1 g) (foldUp (VAL v) d2 g)
    | foldUp (CHOICE (p,d1,d2)) d3 g =
        mergeD p (foldUp d1 d3 g) (foldUp d2 d3 g)
    | foldUp _ _ _ = raise DistribError ("Internal error foldUp",(0,0))

  fun singletons (d as (VAL [n])) pos = d
    | singletons (CHOICE (p,d1,d2)) pos =
         mergeD p (singletons d1 pos) (singletons d2 pos)
    | singletons _ pos = raise DistribError ("Singleton expected",pos)

  fun arith (VAL [m]) (VAL [n]) f pos = VAL [f (m,n)]
    | arith (CHOICE (p, d1, d2)) d3 f pos =
        mergeD p (arith d1 d3 f pos) (arith d2 d3 f pos)
    | arith d1 (CHOICE (p, d2, d3)) f pos =
        mergeD p (arith d1 d2 f pos) (arith d1 d3 f pos)
    | arith _ _ _ pos = raise DistribError ("Internal error arith",pos)

  fun linear (VAL v) f = VAL (f v)
    | linear (CHOICE (p, d1, d2)) f =
        choice (p, linear d1 f, linear d2 f)
    | linear (CPAIR (d1,d2)) f =
        cpair (linear d1 f, linear d2 f)

  fun toList p (VAL v) = [(v,p)]
    | toList p (CHOICE (p1, VAL v, d)) =
        (v,p*p1) :: toList (p*(1.0-p1)) d
    | toList _ _ = raise DistribError ("Internal error toList",(0,0))

  fun distribDice e = toList 1.0 (normalize (dExp e []))

  and dExp exp table =
    case exp of
      Syntax.NUM (n,p) => VAL [n]
    | Syntax.ID (x,p) =>
        (case lookup x table of
           SOME v => VAL v
         | NONE => raise DistribError ("unknown variable: "^x,p))
    | Syntax.PLUS (e1,e2,p) =>
	arith (singletons (dExp e1 table) p)
	      (singletons (dExp e2 table) p)
	      (op +) p
    | Syntax.MINUS (e1,e2,p) =>
	arith (singletons (dExp e1 table) p)
	      (singletons (dExp e2 table) p)
	      (op -) p
    | Syntax.UMINUS (e1,p) =>
	arith (VAL [0])
	      (singletons (dExp e1 table) p)
	      (op -) p
    | Syntax.TIMES (e1,e2,p) =>
	arith (singletons (dExp e1 table) p)
	      (singletons (dExp e2 table) p)
	      (op * ) p
    | Syntax.DIVIDE (e1,e2,p) =>
	arith (singletons (dExp e1 table) p)
	      (singletons (dExp e2 table) p)
	      (op div) p
    | Syntax.CONC (e1,e2,p) =>
        CPAIR (dExp e1 table, dExp e2 table)
    | Syntax.D (e1,p) =>
	let
	  fun d 1 m = VAL [m]
	    | d n m =
	        if n>0 then
		  CHOICE (1.0/Real.fromInt n,VAL [m], d (n-1) (m+1))
		else raise DistribError ("arg to d or D must be >0",p)
	  fun ds (VAL [n]) = d n 1
	    | ds (CHOICE (p,d1,d2)) =
	        mergeD p (ds d1) (ds d2)
	    | ds _ = raise DistribError ("Internal error D",p)
	in
	  ds (singletons (dExp e1 table) p)
	end
    | Syntax.SUM (e1,p) =>
	foldD (dExp e1 table)
	      (fn v => VAL [List.foldr (op +) 0 v])
	      (fn ([m],[n]) => VAL [m+n]
		| _ => raise DistribError ("Internal error SUM",p))
    | Syntax.COUNT (e1,p) =>
	foldD (dExp e1 table)
	      (fn v => VAL [List.length v])
	      (fn ([m],[n]) => VAL [m+n]
		| _ => raise DistribError ("Internal error COUNT",p))
    | Syntax.LEAST (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun least n v =
	        if n<0 then
		  raise DistribError ("arg to least must be >=0",p)
		else if List.length v <= n then VAL v
		else VAL (List.take (v, n))
	  fun ls (VAL [n]) =
	        foldD d2 (least n) (fn (v,w) => least n (mergeI v w))
	    | ls (CHOICE (p,d1,d2)) =
	        mergeD p (ls d1) (ls d2)
	    | ls _ = raise DistribError ("Internal error LEAST",p)
	in
	  ls (singletons (dExp e1 table) p)
	end
    | Syntax.LARGEST (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun largest n v =
	        let val ll = List.length v in
		  if n<0 then
		    raise DistribError ("arg to largest must be >=0",p)
		  else if ll <= n then VAL v
		  else VAL (List.drop (v, ll - n))
		end
	  fun ls (VAL [n]) =
	        foldD d2 (largest n) (fn (v,w) => largest n (mergeI v w))
	    | ls (CHOICE (p,d1,d2)) =
	        mergeD p (ls d1) (ls d2)
	    | ls _ = raise DistribError ("Internal error LARGEST",p)
	in
	  ls (singletons (dExp e1 table) p)
	end
    | Syntax.HASH (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun h 0 d = VAL []
	    | h 1 d = d
	    | h n d = if n<0 then
	                raise DistribError ("first arg to # must be >=0",p)
	              else if n mod 2 = 0 then
			  let val d3 = h (n div 2) d
			  in cpair (d3, d3) end
		      else cpair (d, h (n-1) d)
	  fun hs (VAL [n]) = h n d2
	    | hs (CHOICE (p,d1,d2)) =
	        choice (p, hs d1, hs d2)
	    | hs _ = raise DistribError ("Internal error #",p)
	in
	  hs (singletons (dExp e1 table) p)
	end
    | Syntax.LT (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun lt n = linear d2 (List.filter (fn a => a<n))
	  fun filter (VAL [n]) = lt n
	    | filter (CHOICE (p,d1,d2)) =
	        choice (p, filter d1, filter d2)
	    | filter _ = raise DistribError ("Internal error <",p)
	in
	  filter (singletons (dExp e1 table) p)
	end
    | Syntax.LE (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun le n = linear d2 (List.filter (fn a => a<=n))
	  fun filter (VAL [n]) = le n
	    | filter (CHOICE (p,d1,d2)) =
	        choice (p, filter d1, filter d2)
	    | filter _ = raise DistribError ("Internal error <",p)
	in
	  filter (singletons (dExp e1 table) p)
	end
    | Syntax.GT (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun gt n = linear d2 (List.filter (fn a => a>n))
	  fun filter (VAL [n]) = gt n
	    | filter (CHOICE (p,d1,d2)) =
	        choice (p, filter d1, filter d2)
	    | filter _ = raise DistribError ("Internal error <",p)
	in
	  filter (singletons (dExp e1 table) p)
	end
    | Syntax.GE (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun ge n = linear d2 (List.filter (fn a => a>=n))
	  fun filter (VAL [n]) = ge n
	    | filter (CHOICE (p,d1,d2)) =
	        choice (p, filter d1, filter d2)
	    | filter _ = raise DistribError ("Internal error <",p)
	in
	  filter (singletons (dExp e1 table) p)
	end
    | Syntax.EQ (e1,e2,p) =>
        let
	  val d2 = dExp e2 table
	  fun eq n = linear d2 (List.filter (fn a => a=n))
	  fun filter (VAL [n]) = eq n
	    | filter (CHOICE (p,d1,d2)) =
	        choice (p, filter d1, filter d2)
	    | filter _ = raise DistribError ("Internal error <",p)
	in
	  filter (singletons (dExp e1 table) p)
	end
    | Syntax.FROMTO (e1,e2,p) =>
        let
	  fun ft (VAL [m]) (VAL [n]) =
	        VAL (List.tabulate (n-m+1,fn x=>x+m))
	    | ft (CHOICE (p, d1, d2)) d3 =
		mergeD p (ft d1 d3) (ft d2 d3)
	    | ft d1 (CHOICE (p, d2, d3)) =
	        mergeD p (ft d1 d2) (ft d1 d3)
	    | ft _ _ = raise DistribError ("Internal error FROMTO",p)
	in
	  ft (singletons (dExp e1 table) p)
	     (singletons (dExp e2 table) p)
	end
    | Syntax.LET (x,e1,e2,p) =>
        let
	  fun bind (VAL v) = dExp e2 ((x,v)::table)
	    | bind (CHOICE (p,d1,d2)) =
	        choice (p, bind d1, bind d2)
	    | bind _ = raise DistribError ("Internal error LET ",p)
	in
	  bind (normalize (dExp e1 table))
	end
    | Syntax.REPEAT (x,e1,e2,p) =>
        let
	  fun repeat (d as (VAL [])) n memo = (d, memo)
	    | repeat (d as (VAL v)) n memo =
	        let
		  val (d1,memo1) =
		        repeat0 (normalize (dExp e2 ((x,v)::table)))
				(n-1)
				memo
		in
	          (cpair (d,d1), memo1)
	        end
	    | repeat (CHOICE (p,d1,d2)) n memo =
	        let
		  val (d11,memo1) = repeat0 d1 n memo
		  val (d21,memo2) = repeat0 d2 n memo1
		in
	          (choice (p, d11, d21), memo2)
		end
	    | repeat _ _ _ = raise DistribError ("Internal error REPEAT ",p)
	  and repeat0 d 0 memo = (VAL [], memo)
	    | repeat0 d n memo =
	        case lookup (n,d) memo of
		  NONE => let
		            val (d1,memo1) = repeat d n memo
			  in
			    (d1, ((n,d),d1)::memo1)
			  end
		| SOME d1 => (d1, memo)
	  val d0 = normalize (dExp e1 table)
	in
	  #1 (repeat0 d0 (!maxiterations) [])
	end
    | Syntax.FOREACH (x,e1,e2,p) =>
        let
	  fun foreach [] = VAL []
	    | foreach (n::ns) =
	        cpair (dExp e2 ((x,[n])::table),
		       foreach ns)
          fun foreach0 (VAL v) = foreach v
	    | foreach0 (CHOICE (p,d1,d2)) =
	        choice (p, foreach0 d1, foreach0 d2)
	    | foreach0 _ = raise DistribError ("Internal error foreach ",p)
	in
	  foreach0 (normalize (dExp e1 table))
	end
    | Syntax.IF (e1,e2,e3,p) =>
        let
	  val d1 = dExp e1 table
	  fun pEmpty (VAL []) = 1.0
	    | pEmpty (VAL v) = 0.0
	    | pEmpty (CHOICE (p,d1,d2)) =
	        p*(pEmpty d1)+(1.0-p)*(pEmpty d2)
	    | pEmpty (CPAIR (d1,d2)) =
	        let
		  val p1 = pEmpty d1
		in
		  if p1=0.0 then 0.0
		  else p1*(pEmpty d2)
		end
	  val p1 = pEmpty d1
	in
	  if p1 = 0.0 then dExp e2 table
	  else if p1 = 1.0 then dExp e3 table
	  else choice (p1, dExp e3 table, dExp e2 table)
	end

end

