{

 open Lexing;
 val currentLine = ref 1
 val lineStartPos = ref [0]

 fun getPos lexbuf = getLineCol (getLexemeStart lexbuf)
                                (!currentLine)
                                (!lineStartPos)

 and getLineCol pos line (p1::ps) =
       if pos>=p1 then (line, pos-p1)
       else getLineCol pos (line-1) ps


 exception LexicalError of string * (int * int) (* (message, (line, column)) *)

 fun lexerError lexbuf s = 
     raise LexicalError (s, getPos lexbuf)

 fun keyword (s, pos) =
     case s of
         "D"            => Parser.D pos
       | "d"            => Parser.D pos
       | "sum"          => Parser.SUM pos
       | "least"        => Parser.LEAST pos
       | "largest"      => Parser.LARGEST pos
       | "count"        => Parser.COUNT pos
       | "let"          => Parser.LET pos
       | "in"           => Parser.IN pos
       | "repeat"       => Parser.REPEAT pos
       | "foreach"      => Parser.FOREACH pos
       | "do"           => Parser.DO pos
       | "if"           => Parser.IF pos
       | "then"         => Parser.THEN pos
       | "else"         => Parser.ELSE pos
       | _              => Parser.ID (s,pos);

 }

rule Token = parse
    [` ` `\t` `\r`]     { Token lexbuf } (* whitespace *)
  | [`\n` `\012`]       { currentLine := !currentLine+1;
                          lineStartPos :=  getLexemeStart lexbuf
                                           :: !lineStartPos;
                          Token lexbuf } (* newlines *)
  | "\\" [^ `\n` `\012`]*
                        { Token lexbuf } (* comment *)
  | [`0`-`9`]+          { case Int.fromString (getLexeme lexbuf) of
                               NONE   => lexerError lexbuf "Bad integer"
                             | SOME i => Parser.NUM (i, getPos lexbuf)
                        }
  | ([`a`-`z`] | [`A`-`Z`]) ([`a`-`z`] | [`A`-`Z`])*
                        { keyword (getLexeme lexbuf, getPos lexbuf) }
  | `+`                 { Parser.PLUS (getPos lexbuf) }
  | `-`                 { Parser.MINUS (getPos lexbuf) }
  | `*`                 { Parser.TIMES (getPos lexbuf) }
  | `/`                 { Parser.DIVIDE (getPos lexbuf) }
  | `(`                 { Parser.LPAR (getPos lexbuf) }
  | `)`                 { Parser.RPAR (getPos lexbuf) }
  | `=`                 { Parser.EQ (getPos lexbuf) }
  | `<`                 { Parser.LT (getPos lexbuf) }
  | `>`                 { Parser.GT (getPos lexbuf) }
  | "<="                { Parser.LE (getPos lexbuf) }
  | ">="                { Parser.GE (getPos lexbuf) }
  | ".."                { Parser.DOTDOT (getPos lexbuf) }
  | `@`                 { Parser.CONC (getPos lexbuf) }
  | `#`                 { Parser.HASH (getPos lexbuf) }
  | eof                 { Parser.EOF (getPos lexbuf) }
  | _                   { lexerError lexbuf "Illegal symbol in input" }

;
