structure Main =
struct

  exception ParamErr of string

  val percent = ref true

  val graph = ref 0.0

  fun createLexerStream ( is : BasicIO.instream ) =
      Lexing.createLexer (fn buff => fn n => Nonstdio.buff_input is buff 0 n)

  fun print s = TextIO.output (TextIO.stdOut, s)

  fun stringVal l =
        String.concat
	  (List.map
	    (fn n =>
	      if n>=0 then Int.toString n ^ " "
	      else "-" ^ Int.toString (~n) ^ " ")
	    l)

  fun realtostring n = if n<0.0 then "-" ^ Real.toString (~n)
		       else Real.toString n

  fun printVal l = print (stringVal l ^"\n")

  fun gtList [] _ = false
    | gtList (a::l1) [] = true
    | gtList (a::l1) (b::l2) = a>b orelse a=b andalso gtList l1 l2


  fun printDist1 [] pad acc = ()
    | printDist1 ((a,p)::l) pad acc =
        let
          val s1 = stringVal a
	  val s2 = String.substring(pad,
				    0,String.size pad - String.size s1)
	  val s3 = if p<9.999999
		   then " "^realtostring p
		   else realtostring p
	  val s4 = String.substring("                 ",
				    0,18-String.size s3)
	  val s5 = if acc<9.999999
		   then " "^realtostring acc
		   else realtostring acc
	  val s6 = String.substring("                 ",
				    0,18-String.size s5)
          fun bar p = if (!graph)*p<0.5 then []
                      else #"|" :: bar (p-1.0/(!graph))
	in
          (if !graph=0.0
           then print (s2 ^ s1 ^ ":  " ^ s3 ^ s4 ^ s5 ^ s6 ^ "\n")
           else  print (s2 ^ s1 ^ ":  " ^ implode (bar p) ^ "\n");
	   printDist1 l pad (acc+p))
	end

  fun printDist l =
        let
	  fun myAdd (([],p),s) = s
            | myAdd (((n::l),p),s) = p*(Real.fromInt n)+s
	  fun myAdd2 (([],p),s) = s
            | myAdd2 (((n::l),p),s) = p*(Real.fromInt n)*(Real.fromInt n)+s
	  val maxLen = List.foldr Int.max 0
		        (List.map (String.size o stringVal o (#1)) l)
	  val pad = String.implode (List.tabulate (maxLen+5, fn x => #" "))
	  val s1 = "Value"
	  val s2 = String.substring(pad,
				    0,String.size pad - String.size s1)
	  val s3 = if !percent
	           then "    % =             "
		   else "  Probability for = "
	  val s4 = if !percent
	           then "  % <             "
		   else " Probability for <"
	  val mean = if List.all (fn x => List.length (#1 x)<=1) l
		     then SOME (List.foldr myAdd 0.0 l,
			        List.foldr myAdd2 0.0 l)
		     else NONE
	  val s5 = case mean of
		     NONE => ""
		   | SOME (m,m2) =>
		       "\nAverage = " ^ realtostring m ^
		       "    Spread = " ^
		       realtostring (Math.sqrt(m2-m*m)) ^ "\n"
	  val l1 = if !percent
	           then List.map (fn (a,p)=>(a,100.0*p)) l
		   else l
        in
	  (if !graph=0.0
           then print (s2 ^ s1 ^ s3 ^ s4 ^ "\n")
           else print (s2 ^ s1 ^"   "^ Real.toString (!graph) ^" bars per %\n");
	   printDist1 l1 pad 0.0;
	   print s5)
	end

  fun run filename n defs =
        let
          val lexbuf =  createLexerStream
			  (if filename = "" then BasicIO.std_in
		           else BasicIO.open_in filename)
          val dice = defs (Parser.Dice Lexer.Token lexbuf)
        in
          (let
	     prim_val getrealtime_ : unit -> {sec : int, usec : int} 
                                = 1 "sml_getrealtime"
             val {sec, usec} = getrealtime_ ()
	   in
	     (Interpreter.seed := sec mod 100001 + 5*(usec mod 77771);
	      Interpreter.seed2 := sec mod 3557 + 7*(usec mod 3307))
	   end;
	   if n<0 then (Distribution.maxiterations := ~n;
		        printDist (Distribution.distribDice dice))
	   else if n=0 then printDist (Distribution.distribDice dice)
	   else
	     hd (List.tabulate
	         (n,fn x => printVal (Interpreter.rollDice dice)))
	  )
        end

  fun errorMess s = TextIO.output (TextIO.stdErr,s ^ "\n");

  fun findDef str =
        case String.fields (fn c=>c = #"=") str of
          [name,valString] =>
            (case Int.fromString valString of
	       NONE => NONE
	     | SOME value => SOME (name,value))
        | _ => NONE

  fun run0 [] filename n defs = run filename n defs
    | run0 (arg::args) filename n defs =
	case Int.fromString arg of
	  NONE => (case findDef arg of
		     NONE =>
		       if arg = "-p"
		       then (percent := not (!percent);
			     run0 args filename n defs)
		       else if "-g" <= arg andalso arg < "-h"
		       then (case Real.fromString
                                    (String.extract (arg,2,NONE)) of
                               NONE => raise ParamErr arg
                             | SOME s =>
                               (graph := s;
			        run0 args filename n defs))
		       else run0 args arg n defs
		   | SOME (name,value) =>
		       run0 args filename n
			    (fn d=>Syntax.LET (name,
				    Syntax.NUM (value,(0,0)),
				    defs d,(0,0))))
	| SOME n => run0 args filename n defs

  val _ =
    let
      val argv = Mosml.argv ()
    in
      run0 (tl argv) "" 1 (fn d=>d)
    end
    handle Parsing.yyexit ob => errorMess "Parser-exit\n"
         | Parsing.ParseError ob =>
             let val Location.Loc (p1,p2)
                       = Location.getCurrentLocation ()
                 val (lin,col)
                        = Lexer.getLineCol p2
                                          (!Lexer.currentLine)
                                          (!Lexer.lineStartPos)

             in
               errorMess ("Parse-error at line "
                ^ makestring lin ^ ", column " ^ makestring col)
             end
         | Lexer.LexicalError (mess,(lin,col)) =>
               errorMess ("Lexical error: " ^mess^ " at line "
                ^ makestring lin ^ ", column " ^ makestring col)
	 | Interpreter.RunError (mess,(lin,col)) =>
               errorMess ("Runtime error: " ^mess^ " at line "
                ^ makestring lin ^ ", column " ^ makestring col)
	 | Distribution.DistribError (mess,(lin,col)) =>
               errorMess ("Distribution error: " ^mess^ " at line "
                ^ makestring lin ^ ", column " ^ makestring col)
         | SysErr (s,_) => errorMess ("Exception: " ^ s)
         | ParamErr s => errorMess ("Bad command-line parameter: " ^ s)
end

